// https://www.redblobgames.com/grids/hexagons/

// Horizontal distance between hex centers is sqrt(3) * size. The vertical
// distance is 3 / 2 * size. When we calculate horzDist / vertDist, the size
// cancels out, leaving us with a unitless ratio of sqrt(3) / (3 / 2), or
// 2 * sqrt(3) / 3.

const svgns = "http://www.w3.org/2000/svg",
  horzToVertDistRatio = 2 * Math.sqrt(3) / 3,

  arcSize = {
    'small': Math.atan(horzToVertDistRatio / 6),
    'medium': Math.atan(horzToVertDistRatio / 2),
    'large': Math.atan(7 * horzToVertDistRatio / 2)
  },

  firingArcVisibility = {
    davion: false,
    liao: false
  },

  clippedFiringArcRadius = 25;

let svg;

function calculateAngle(xDiff, yDiff) {
  yDiff = -yDiff;
  let angle = Math.abs(Math.atan(yDiff / xDiff));

  if (xDiff < 0 && yDiff > 0) {
    angle = Math.PI - angle;
  } else if (xDiff < 0 && yDiff < 0) {
    angle = Math.PI + angle;
  } else if (xDiff > 0 && yDiff < 0) {
    angle = 2 * Math.PI - angle;
  }

  return angle;
}

function edgePoint({ x: x1, y: y1 }, { x: x2, y: y2 }, { x: [minX, maxX], y: [minY, maxY] }) {
  const xDiff = x2 - x1,
    yDiff = y2 - y1,
    xIntercept = y => (y - y1) * xDiff / yDiff + x1,
    yIntercept = x => (x - x1) * yDiff / xDiff + y1;

  let pointCoords;

  if (xDiff > 0 && yDiff > 0) {
    let x = xIntercept(maxY);

    pointCoords = x <= maxX ? [x, maxY] : [maxX, yIntercept(maxX)];
  } else if (xDiff > 0 && yDiff < 0) {
    let y = yIntercept(maxX);

    pointCoords = y >= minY ? [maxX, y] : [xIntercept(minY), minY];
  } else if (xDiff < 0 && yDiff < 0) {
    let x = xIntercept(minY);

    pointCoords = x >= minX ? [x, minY] : [minX, yIntercept(minX)];
  } else {
    let y = yIntercept(minX);

    pointCoords = y <= maxY ? [minX, y] : [xIntercept(maxY), maxY];
  }

  return new Point(...pointCoords);
}

class Point {
  constructor(x = 0, y = 0) {
    this.x = +x;
    this.y = +y;
  }

  toString() {
    return `${this.x},${this.y}`;
  }
}

function position(e, firingArc, firingArcOutline, aimLine, grid) {
  // TODO: handle exactly horizontal and vertical lines?

  let pointer = new DOMPoint(e.clientX, e.clientY);
  let pointerPt = pointer.matrixTransform(grid.getScreenCTM().inverse());

  let pivotPt = new Point(aimLine.getAttribute('x1'), aimLine.getAttribute('y1'));

  let { x, y, width, height } = grid.getBBox();

  const bounds = {
    x: [x, x + width],
    y: [y, y + height]
  }

  let cornerPoints = [
    [x, y],
    [x + width, y],
    [x + width, y + height],
    [x, y + height]
  ].map(([x, y]) => new Point(x, y));

  let aimPt = edgePoint(pivotPt, pointerPt, bounds);
  aimLine.setAttributeNS(null, 'x2', aimPt.x);
  aimLine.setAttributeNS(null, 'y2', aimPt.y);

  let angle = calculateAngle(aimPt.x - pivotPt.x, aimPt.y - pivotPt.y);
  let arcAngle = arcSize[firingArc.dataset.size];
  let distance = Math.sqrt((aimPt.x - pivotPt.x) ** 2 + (aimPt.y - pivotPt.y) ** 2);
  let yDelta = distance * Math.cos(angle) * Math.tan(arcAngle);
  let xDelta = distance * Math.sin(angle) * Math.tan(arcAngle);

  let arcPt1 = new Point(aimPt.x - xDelta, aimPt.y - yDelta);
  let arcPt2 = new Point(aimPt.x + xDelta, aimPt.y + yDelta);

  arcPt1 = edgePoint(pivotPt, arcPt1, bounds);
  arcPt2 = edgePoint(pivotPt, arcPt2, bounds);

  firingArcOutline.setAttributeNS(null, 'points', `${arcPt1} ${pivotPt} ${arcPt2}`);

  function touchSameEdge({ x: x1, y: y1 }, { x: x2, y: y2 }) {
    return x1 === x2 || y1 === y2;
  }

  function touchOrthogonalEdges({ x: x1, y: y1 }, { x: x2, y: y2 }) {
    return (bounds.x.includes(x1) && bounds.y.includes(y2)) || (bounds.y.includes(y1) && bounds.x.includes(x2));
  }

  function shareValue({ x: x1, y: y1 }, { x: x2, y: y2 }) {
    return x1 === x2 || y1 === y2;
  }

  // which arcpt does aimpt share a value with?
  // if they share an x value, we will look for corner y value
  // if they share a y value, we will look for corner x value
  // is aim pt non-shared value greater or less than arcpt non-shared value?
  function findWhichCorners(pt, ...pts) {
    const ptVals = Object.values(pt),
      sharedValPt = pts.find(({ x, y }) => ptVals.includes(x) || ptVals.includes(y));

    if (!sharedValPt) {
      return;
    }

    const nonSharedValKey = pt.x === sharedValPt.x ? 'x' : 'y';
    let cornerVal;

    if (pt[nonSharedValKey] < sharedValPt[nonSharedValKey]) {
      cornerVal = Math.min(...bounds[nonSharedValKey]);
    } else {
      cornerVal = Math.max(...bounds[nonSharedValKey]);
    }

    return cornerPoints.filter(cp => cp[nonSharedValKey] === cornerVal);
  }

  let points;

  if (touchSameEdge(arcPt1, arcPt2)) {
    // 0-corner case
    points = [arcPt2, pivotPt, arcPt1];
  } else if (touchOrthogonalEdges(arcPt1, arcPt2)) {
    if (touchSameEdge(aimPt, arcPt1) || touchSameEdge(aimPt, arcPt2)) {
      // 1-corner case
      let cp = cornerPoints.find(cp => shareValue(cp, arcPt1) && shareValue(cp, arcPt2));
      points = [arcPt2, pivotPt, arcPt1, cp];
    } else {
      // 3-corner case
      points = cornerPoints.filter(cp => !shareValue(cp, arcPt1) || !shareValue(cp, arcPt2));
      let index = points.findIndex(cp => shareValue(cp, arcPt2));
      points.splice(index + 1, 0, arcPt2, pivotPt, arcPt1);
    }
  } else {
    if (touchSameEdge(aimPt, arcPt1) || touchSameEdge(aimPt, arcPt2)) {
      // 2-corner case, aim and an arc point touch the same edge
      points = findWhichCorners(aimPt, arcPt1, arcPt2);
      let index = points.findIndex(cp => shareValue(cp, arcPt2));
      points.splice(index + 1, 0, arcPt2, pivotPt, arcPt1);
    } else {
      // 2-corner case, aim and both arc points all touch different edges
      points = cornerPoints.filter(cp => shareValue(cp, aimPt) || shareValue(cp, aimPt));
      let index = points.findIndex(cp => shareValue(cp, arcPt2));
      points.splice(index + 1, 0, arcPt2, pivotPt, arcPt1);
    }
  }

  firingArc.setAttributeNS(null, 'points', points.join(' '));
}

function setDataAttrs({ dataset: { allegiance, number }}, el) {
  el.dataset.allegiance = allegiance;
  el.dataset.number = number;
}

function getClipPathId({ dataset: { allegiance, number }}) {
  return `clip-path-${allegiance}-${number}`;
}

function getUnclipped() {
  return svg.querySelectorAll('#firing-arcs :not([clip-path])');
};

export default function (el) {
  svg = el;

  this.set = function (size, counter, { x, y }) {
    this.get(counter).forEach(fa => fa.remove());

    let arcLayer = svg.querySelector('#shapes');
    let outlineLayer = svg.querySelector('#lines');
    let arcContainer = svg.querySelector('#firing-arcs');
    let grid = svg.querySelector('.grid');

    let aimLine = document.createElementNS(svgns, 'line');
    aimLine.setAttributeNS(null, 'x1', x);
    aimLine.setAttributeNS(null, 'y1', y);
    aimLine.setAttributeNS(null, 'x2', x);
    aimLine.setAttributeNS(null, 'y2', y);
    outlineLayer.appendChild(aimLine);

    let firingArc = document.createElementNS(svgns, 'polygon');
    setDataAttrs(counter, firingArc);
    firingArc.setAttributeNS(null, 'points', `${x},${y}`);
    firingArc.dataset.size = size;
    firingArc.classList.add('firing-arc', 'active');

    let firingArcOutline = document.createElementNS(svgns, 'polyline');
    setDataAttrs(counter, firingArcOutline);
    firingArcOutline.setAttributeNS(null, 'points', `${x},${y}`);

    let clipShape = document.createElementNS(svgns, 'circle');
    clipShape.setAttributeNS(null, 'cx', x);
    clipShape.setAttributeNS(null, 'cy', y);
    clipShape.setAttributeNS(null, 'r', clippedFiringArcRadius);

    let clipPath = document.createElementNS(svgns, 'clipPath');
    setDataAttrs(counter, clipPath);
    clipPath.setAttributeNS(null, 'id', getClipPathId(counter));
    clipPath.appendChild(clipShape);

    arcContainer.appendChild(clipPath);
    arcLayer.appendChild(firingArc);
    outlineLayer.appendChild(firingArcOutline);

    // const positionListener = position.bind(svg, firingArc, firingArcOutline, aim, grid);
    function positionListener(e) {
      position(e, firingArc, firingArcOutline, aimLine, grid);
    }

    let placementListener = e => {
      grid.removeAttribute('style');
      aimLine.remove();
      svg.querySelectorAll('.firing-arc.active').forEach(el => el.classList.remove('active'));
      svg.removeEventListener('mousemove', positionListener);
    };

    let cancelPlacementListener = e => {
      e.preventDefault();

      this.get(counter).forEach(fa => fa.remove());
      grid.removeAttribute('style');
      svg.removeEventListener('mousemove', positionListener);
    };

    grid.style.pointerEvents = 'none';
    firingArc.addEventListener('click', placementListener, { once: true });
    firingArc.addEventListener('contextmenu', cancelPlacementListener, { once: true });
    svg.addEventListener('mousemove', positionListener);
  };

  this.clear = function (allegiance) {
    const selector = `#firing-arcs [data-allegiance="${allegiance}"]`;
    svg.querySelectorAll(selector).forEach(el => el.remove());
  };

  this.get = function ({ dataset: { allegiance, number }}) {
    return svg.querySelectorAll(`#firing-arcs [data-number="${number}"][data-allegiance="${allegiance}"], #firing-arcs line`);
  };

  this.toggleVisibility = function (allegiance) {
    const vis = firingArcVisibility[allegiance],
      clipPaths = svg.querySelectorAll(`clipPath[data-allegiance="${allegiance}"]`);

    clipPaths.forEach(cp => cp.style.display = !vis ? 'none' : '');
    firingArcVisibility[allegiance] = !vis;
  };

  this.toggleCounterVisibility = function ({ dataset: { number, allegiance }}, vis) {
    const cp = svg.querySelector(`#clip-path-${allegiance}-${number}`),
      display = vis ? 'none' : '';

    if (cp) {
      cp.style.display = firingArcVisibility[allegiance] ? 'none' : display;
    }
  };

  this.clipAll = function () {
    let unclipped = getUnclipped();

    unclipped.forEach(el => {
      const { number, allegiance } = el.dataset,
        clipPathId = `clip-path-${allegiance}-${number}`,
        isVisible = firingArcVisibility[allegiance];

      if (isVisible) {
        svg.querySelector(`#${clipPathId}`).style.display = 'none';
      }

      el.setAttributeNS(null, 'clip-path', `url(#${clipPathId})`);
    });
  };
}
